%--------------------------------------------------------------------
% Numerical Example of LPV robust systhesis using IQCs
% Update: 04/02/2015
% Shu Wang
%--------------------------------------------------------------------
clear all;
clc;
%% Frho and weighting functions

rho = [1:0.5:3];  % time varying parameter rho (5 gridding points)
% rho = [1:0.25:3];  % time varying parameter rho (11 gridding points)

Np=size(rho,2);     % number of gridding points

tau = 71 + 2.*rho;  % system parameter tau

K11 = 87 + 0.2.*rho.^2;         % system parameter K
K12 = -87.2 + 0.2.*rho.^2;      % system parameter K
K21 = 107.4 + 0.2.*rho.^2;      % system parameter K
K22 = -110.4+ 0.2.*rho.^2;      % system parameter K

Frho = ss;
for i=1:Np
    
    Ki = [K11(i) K12(i);K21(i) K22(i)];
    Agi = blkdiag(-1/tau(i),-1/tau(i));
    Bgi = 1/tau(i)*eye(2);
    Cgi = Ki;
    Dgi = zeros(2,2);

    Frho(:,:,i) = ss(Agi,Bgi,Cgi,Dgi);
end

AF = Frho.a;
BF = Frho.b;
CF = Frho.c;
DF = Frho.d;

dz1 = 0.05;     % dead zone range
dz2 = 0.05;     % dead zone range

We = 0.3*ss(tf([1 0.1],[2 10^(-5)])*eye(2));    % weight for tracking error

Wu = ss(tf([1 10],[1 100])*eye(2));     % weight for control input

beta1 = 0.6;
beta2 = 0.3;
Wd = ss(blkdiag(beta1,beta2));  % weight for uncertainty

%% Construct the nominal model Gicn

systemnames = 'Frho Wu We';
inputvar = '[r1; r2; u1; u2]';
outputvar = '[ We; Wu; r1 - Frho(1);r2 - Frho(2)]';
input_to_We = '[r1 - Frho(1);r2 - Frho(2)]';
input_to_Wu = '[u1;u2]';
input_to_Frho = '[u1;u2]';
cleanupsysic = 'yes';
Gicn = sysic;

%% nominal LPV controller synthesis 

opt = lpvsynOptions;
opt.BackOffFactor = 1.2;
[Knom,GAMnom,INFOnom] = lpvsyneng(Gicn,2,2);

fprintf('Induced L_2 gain of Fl(Gicn,Knom): %f \t \n', GAMnom)

AKnom = Knom.a;
BKnom = Knom.b;
CKnom = Knom.c;
DKnom = Knom.d;

%% Construct the uncertain model 

systemnames = 'Frho Wu We Wd';
inputvar = '[d1; d2; r1; r2; u1; u2]';
outputvar = '[ Wd(1); Wd(2); We; Wu; r1 - Frho(1);r2 - Frho(2)]';
input_to_We = '[r1 - Frho(1); r2 - Frho(2)]';
input_to_Wu = '[u1; u2]';
input_to_Wd = '[u1; u2]';
input_to_Frho = '[u1 - d1; u2 - d2]';
cleanupsysic = 'yes';
Gic = sysic;

%% Specify IQC multipliers for the dead zone

% norm bound multiplier
Psi_a = ss(eye(2));

% sector bound multiplier
Pi_b = [0 1; 1 -2];
Psi_b = jspec(Pi_b);

% Zames-Falb  multiplier for monotonic odd nonlinearity
Hsat = ss(tf(1,[1 1]));
Pi_c = [0, (1 + Hsat); (1 + Hsat'), -(2 + Hsat + Hsat')];
Psi_c = jspec(Pi_c);

% extended IQCs for block diagonal structured dead zones

Psis{1} = blkdiag(Psi_a,Psi_a);

Psi2_11 = blkdiag(Psi_a(1,1),Psi_b(1,1));
Psi2_12 = blkdiag(Psi_a(1,2),Psi_b(1,2));
Psi2_21 = blkdiag(Psi_a(2,1),Psi_b(2,1));
Psi2_22 = blkdiag(Psi_a(2,2),Psi_b(2,2));
Psis{2} = [Psi2_11 Psi2_12;Psi2_21 Psi2_22];

Psi3_11 = blkdiag(Psi_b(1,1),Psi_a(1,1));
Psi3_12 = blkdiag(Psi_b(1,2),Psi_a(1,2));
Psi3_21 = blkdiag(Psi_b(2,1),Psi_a(2,1));
Psi3_22 = blkdiag(Psi_b(2,2),Psi_a(2,2));
Psis{3} = [Psi3_11 Psi3_12;Psi3_21 Psi3_22];

Psi4_11 = blkdiag(Psi_a(1,1),Psi_c(1,1));
Psi4_12 = blkdiag(Psi_a(1,2),Psi_c(1,2));
Psi4_21 = blkdiag(Psi_a(2,1),Psi_c(2,1));
Psi4_22 = blkdiag(Psi_a(2,2),Psi_c(2,2));
Psis{4} = [Psi4_11 Psi4_12;Psi4_21 Psi4_22];

Psi5_11 = blkdiag(Psi_c(1,1),Psi_a(1,1));
Psi5_12 = blkdiag(Psi_c(1,2),Psi_a(1,2));
Psi5_21 = blkdiag(Psi_c(2,1),Psi_a(2,1));
Psi5_22 = blkdiag(Psi_c(2,2),Psi_a(2,2));
Psis{5} = [Psi5_11 Psi5_12;Psi5_21 Psi5_22];

%% Initialization

szDelta = [2 2];
szK = [2 2];

szG = size(Gic);
if numel(szG) == 2
    szG = [szG 1];
end
nmod = szG(3); % # of models in the G array.

szU = sum(szDelta,1);
nw = szU(1);   % # of outputs of Delta (uncertaint inputs to G)
nv = szU(2);   % # of inputs to Delta

nu = szK(1);    % # of outputs of K (control inputs)
ny = szK(2);    % # of inputs of K (measurements)

nd = szG(2) - nw - nu ;   % # of disturbances
ne = szG(1) - nv - ny ;   % # of errors

% n IQCs
niqcs = numel(Psis);

% lambda for n IQCs
lam = zeros(niqcs,1);
lam(1) = 1;

M = blkdiag(eye(nv),-eye(nw));

% stopping tolerence
maxiter = 25;

gamold = inf;
gamtol = 0.05;

%% Iterations

tic;

for ii = 1:maxiter

    fprintf('\n Iteration: %i \n', ii);
    
    % create a single IQC from n IQCs
    if ii ~= 1
        Pi = ss(zeros(nv + nw,nv + nw));
        Scl = blkdiag(1/gamold*eye(nv),eye(nw));
        for jj = 1:niqcs
            Pi = Pi + lam(jj)*Scl*Psis{jj}'*M*Psis{jj}*Scl;
        end
    end
    
    % facotorization of the single IQC
    if ii == 1
        Psi = Psis{1};
    else
        [Psi,M] = jspec(Pi); 
        invScl = blkdiag(gamold*eye(nv),eye(nw));
        Psi = invScl*Psi;
    end

    % create Psi_plus by inverting w/w_tilde channels
    Apsi = Psi.a;
    Bpsi = Psi.b;
    Cpsi_vtil = Psi.c(1:nv,:);
    Cpsi_wtil = Psi.c(nv+1:nv+nw,:);
    Dpsi_vtil = Psi.d(1:nv,:);
    Dpsi_wtilv = Psi.d(nv+1:nv+nw,1:nv);
    Dpsi_wtilw = Psi.d(nv+1:nv+nw,nv+1:nv+nw);    

    nxpsi = size(Apsi,1);
    
    Psi_mtx = [Apsi, Bpsi; Cpsi_vtil, Dpsi_vtil; zeros(nw,nxpsi+nv), eye(nw)];


    T = [eye(nxpsi), zeros(nxpsi,nw+nv); zeros(nv,nxpsi+nw), eye(nv); 
            -Dpsi_wtilw\Cpsi_wtil, inv(Dpsi_wtilw), -Dpsi_wtilw\Dpsi_wtilv];

    Psi_plusmtx = Psi_mtx*T;
    
    Psi_plusa = Psi_plusmtx(1:nxpsi,1:nxpsi);
    Psi_plusb = Psi_plusmtx(1:nxpsi,nxpsi+1:nxpsi+nw+nv);
    Psi_plusc = Psi_plusmtx(nxpsi+1:nxpsi+nv+nw,1:nxpsi);
    Psi_plusd = Psi_plusmtx(nxpsi+1:nxpsi+nv+nw,nxpsi+1:nxpsi+nw+nv);
    
    Psi_plus = ss(Psi_plusa,Psi_plusb,Psi_plusc,Psi_plusd);
    
    
    % create scaled system G_scl by lft of G and Psi_plus 
    G_scl = lft(Psi_plus,Gic,nv,nw);

    % nominal LPV synthesis step 
    % this "backs off" the optimal gamma by BackOffFactor = 1.2000
    opt = lpvsynOptions;
    opt.BackOffFactor = 1.2;
    [Krob,GAMrob,INFOrob] = lpvsyneng(G_scl,ny,nu);
   
    fprintf('Gamma controller step: %f \t \n', GAMrob)
    
    % robust performance analysis step
    % get analysis interconnection (replace G by Pic for demo)
    Nic = lft(Gic,Krob);
    
    % options for bisection algorithm
    gamlb = 0;  %lower bound
    
    if ii==1    %upper bound   
        gamub = GAMrob; 
    else
        gamub = GAMrob;
    end

    tol = 0.05;
    
    while gamub-gamlb>tol
        gamtry = (gamlb+gamub)/2;
        
        %run analysis
        [t,info] = lpvrbperf(Nic,szDelta,Psis,gamtry);
        %new bounds for bisection
        if t <= 0
            gamub = gamtry;
        else
            gamlb = gamtry;
        end
    end
    
    gamopt = gamub;
    % run analysis using gamopt
    [t,info] = lpvrbperf(Nic,szDelta,Psis,gamopt);

    fprintf('Gamma analysis step: %f \t \n', gamopt)
    %% save results for next iteration
    lam = info.lam;   
    gamvec(ii) = gamopt;
    
    if gamold-gamopt < gamtol
        break;
    end
    gamold = gamopt;
    
end

toc;

AKrob = Krob.a;
BKrob = Krob.b;
CKrob = Krob.c;
DKrob = Krob.d;


%% Check the nominal performance of the robust controller

Hrob = lft(Gicn,Krob);
[GAMHrob,info] = lpvnorm(Hrob);

fprintf('\n Induced L_2 gain of Fl(Gicn,Krob): %f \t \n', GAMHrob)


%% Check the robust performance of the nominal controller

Hic = lft(Gic,Knom);

% options for bisection algorithm
GAMnomlb = GAMnom;  %lower bound
GAMnomub = 20*GAMnom;

tol = 0.05; %tolerance for bisection

while GAMnomub - GAMnomlb > tol
    GAMnomtry = (GAMnomlb + GAMnomub)/2;

    %run analysis
    [tnom,info] = lpvrbperf(Hic,szDelta,Psis,GAMnomtry);
    %new bounds for bisection
    if tnom <= 0
        GAMnomub = GAMnomtry;
    else
        GAMnomlb = GAMnomtry;
    end
end

GAMnomopt = GAMnomub;
% run analysis using gamopt
[tnom,info] = lpvrbperf(Hic,szDelta,Psis,GAMnomopt);

fprintf('\n Robust performance with K_nom: %f \t \n', GAMnomopt)



