%--------------------------------------------------------------------
% Numerical Example of LPV robust systhesis using line search
% Update: 04/02/2015
% Shu Wang
%--------------------------------------------------------------------
clear all;
clc;
%% Model Construction

rho = [1:0.5:3];  % time varying parameter rho
Np=size(rho,2);

tau = 71 + 2.*rho;  % system parameter tau

K11 = 87 + 0.2.*rho.^2;   % system parameter K
K12 = -87.2 + 0.2.*rho.^2;   % system parameter K
K21 = 107.4 + 0.2.*rho.^2;   % system parameter K
K22 = -110.4+ 0.2.*rho.^2;   % system parameter K

Frho = ss;
for i=1:Np
    
    Ki = [K11(i) K12(i);K21(i) K22(i)];
    Agi = blkdiag(-1/tau(i),-1/tau(i));
    Bgi = 1/tau(i)*eye(2);
    Cgi = Ki;
    Dgi = zeros(2,2);

    Frho(:,:,i) = ss(Agi,Bgi,Cgi,Dgi);
end

AG = Frho.a;
BG = Frho.b;
CG = Frho.c;
DG = Frho.d;

dz1 = 0.05;
dz2 = 0.05;

We = 0.3*ss(tf([1 0.1],[2 10^(-5)])*eye(2));

Wu = ss(tf([1 10],[1 100])*eye(2)); % control input

beta1 = 0.6;
beta2 = 0.3;
Wd = ss(blkdiag(beta1,beta2));

%% Model Construction

systemnames = 'Frho Wu We Wd';
inputvar = '[d1; d2; r1; r2; u1; u2]';
outputvar = '[ Wd(1); Wd(2); We; Wu; r1 - Frho(1);r2 - Frho(2)]';
input_to_We = '[r1 - Frho(1); r2 - Frho(2)]';
input_to_Wu = '[u1; u2]';
input_to_Wd = '[u1; u2]';
input_to_Frho = '[u1 - d1; u2 - d2]';
cleanupsysic = 'yes';
Gic = sysic;

%% specify the IQC multiplier

Psis{1} = ss(eye(4));

%% Initialization

szDelta = [2 2];
szK = [2 2];

szG = size(Gic);
if numel(szG) == 2
    szG = [szG 1];
end
nmod = szG(3); % # of models in the G array.

szU = sum(szDelta,1);
nw = szU(1);   % # of outputs of Delta (uncertaint inputs to G)
nv = szU(2);   % # of inputs to Delta

nu = szK(1);    % # of outputs of K (control inputs)
ny = szK(2);    % # of inputs of K (measurements)

nd = szG(2) - nw - nu ;   % # of disturbances
ne = szG(1) - nv - ny ;   % # of errors

% n IQCs
niqcs = numel(Psis);

lam = [0.02:0.02:100];

M = blkdiag(eye(nv),-eye(nw));

% stopping tolerence
maxiter = 100;

gamopt =zeros(maxiter,1);

gamold = inf;
gamtol = 0.05;

%% Iterations
tic;
for ii = 1:maxiter

    fprintf('\n Iteration: %i \n', ii)
    fprintf(' lambda is: %f \t \n', lam(ii))
    % create a single IQC from n IQCs

    Pi = lam(ii)*Psis{1}'*M*Psis{1};

    % facotorization of the single IQC
    [Psi,M] = jspec(Pi); 

    % create Psi_plus by inverting w/w_tilde channels
    Apsi = Psi.a;
    Bpsi = Psi.b;
    Cpsi_vtil = Psi.c(1:nv,:);
    Cpsi_wtil = Psi.c(nv+1:nv+nw,:);
    Dpsi_vtil = Psi.d(1:nv,:);
    Dpsi_wtilv = Psi.d(nv+1:nv+nw,1:nv);
    Dpsi_wtilw = Psi.d(nv+1:nv+nw,nv+1:nv+nw);    

    nxpsi = size(Apsi,1);
    
    Psi_mtx = [Apsi, Bpsi; Cpsi_vtil, Dpsi_vtil; zeros(nw,nxpsi+nv), eye(nw)];

    T = [eye(nxpsi), zeros(nxpsi,nw+nv); zeros(nv,nxpsi+nw), eye(nv); 
            -Dpsi_wtilw\Cpsi_wtil, inv(Dpsi_wtilw), -Dpsi_wtilw\Dpsi_wtilv];

    Psi_plusmtx = Psi_mtx*T;
    
    Psi_plusa = Psi_plusmtx(1:nxpsi,1:nxpsi);
    Psi_plusb = Psi_plusmtx(1:nxpsi,nxpsi+1:nxpsi+nw+nv);
    Psi_plusc = Psi_plusmtx(nxpsi+1:nxpsi+nv+nw,1:nxpsi);
    Psi_plusd = Psi_plusmtx(nxpsi+1:nxpsi+nv+nw,nxpsi+1:nxpsi+nw+nv);
    
    Psi_plus = ss(Psi_plusa,Psi_plusb,Psi_plusc,Psi_plusd);
    
    %%create scaled system G_scl by lft of Gic and Psi_plus 
    
    G_scl = lft(Psi_plus,Gic,nv,nw); %(replace G by Pic for demo)

    % nominal LPV synthesis step 
    % this "backs off" the optimal gamma by BackOffFactor = 1.2000
    opt = lpvsynOptions;
    opt.BackOffFactor = 1.2;
    [Krob,GAMrob,INFOrob] = lpvsyneng(G_scl,ny,nu);
   
    fprintf('Gamma controller step: %f \t \n', GAMrob)
    
    % robust performance analysis step
    %get analysis interconnection (replace G by Pic for demo)
    Nic = lft(Gic,Krob);
    
    % options for bisection algorithm
    gamlb = 0;  %lower bound
    
    if ii==1    %upper bound   
        gamub = GAMrob; 
    else
%         gamub = gamold; 
        gamub = GAMrob;
    end

    tol = 0.05;
    
    while gamub-gamlb>tol
        gamtry = (gamlb+gamub)/2;
        
        %run analysis
        [t,info] = lpvrbperf(Nic,szDelta,Psis,gamtry);
        %new bounds for bisection
        if t <= 0
            gamub = gamtry;
        else
            gamlb = gamtry;
        end
    end
    
    gamopt(ii) = gamub;
    % run analysis using gamopt
    [t,info] = lpvrbperf(Nic,szDelta,Psis,gamopt(ii));

    fprintf('Gamma analysis step: %f \t \n', gamopt(ii))
    % save results for next iteration
    lamanalysis = info.lam;  
    fprintf('lambda after analysis is: %f \t \n', lamanalysis)
    
end
toc;
