%--------------------------------------------------------------------
% Numerical Example of LTI robust synthesis using IQCs approach and DK
% Updated: 04/02/2015
% Shu Wang
%--------------------------------------------------------------------
clear all;
clc;
%% Frho (rho = 2) and weighting functions

% rho = [1:0.5:3];  % time varying parameter rho
rho = 2;

Np=size(rho,2);

tau = 71 + 2.*rho;  % system parameter tau

K11 = 87 + 0.2.*rho.^2;   % system parameter K
K12 = -87.2 + 0.2.*rho.^2;   % system parameter K
K21 = 107.4 + 0.2.*rho.^2;   % system parameter K
K22 = -110.4+ 0.2.*rho.^2;   % system parameter K

Frho = ss;
for i=1:Np
    
    Ki = [K11(i) K12(i);K21(i) K22(i)];
    Agi = blkdiag(-1/tau(i),-1/tau(i));
    Bgi = 1/tau(i)*eye(2);
    Cgi = Ki;
    Dgi = zeros(2,2);

    Frho(:,:,i) = ss(Agi,Bgi,Cgi,Dgi);
end

AF = Frho.a;
BF = Frho.b;
CF = Frho.c;
DF = Frho.d;

We = 0.3*ss(tf([1 0.1],[2 10^(-5)])*eye(2));    % weight for tracking error

Wu = ss(tf([1 10],[1 100])*eye(2));     % weight for control input

beta1 = 0.6;
beta2 = 0.3;
Wd = ss(blkdiag(beta1,beta2));  % weight for uncertainty

%% Construct the model

systemnames = 'Frho Wu We Wd';
inputvar = '[d1; d2; r1; r2; u1; u2]';
outputvar = '[ Wd(1); Wd(2); We; Wu; r1 - Frho(1);r2 - Frho(2)]';
input_to_We = '[r1 - Frho(1); r2 - Frho(2)]';
input_to_Wu = '[u1; u2]';
input_to_Wd = '[u1; u2]';
input_to_Frho = '[u1 - d1; u2 - d2]';
cleanupsysic = 'yes';
Gic = sysic;
%% DK synthesis 

Un1 = ultidyn('Un1',[1 1]);
Un2 = ultidyn('Un2',[1 1]);

systemnames = 'Frho Wu We Wd Un1 Un2';
inputvar = '[r1; r2; u1; u2]';
outputvar = '[We; Wu; r1 - Frho(1);r2 - Frho(2)]';
input_to_We = '[r1 - Frho(1); r2 - Frho(2)]';
input_to_Wu = '[u1; u2]';
input_to_Wd = '[u1; u2]';
input_to_Un1 = '[Wd(1)]';
input_to_Un2 = '[Wd(2)]';
input_to_Frho = '[u1 - Un1; u2 - Un2]';
cleanupsysic = 'yes';
Gicun = sysic;

tic;
[Kdk,clpdk,RPdk] = dksyn(Gicun,2,2);
toc;

fprintf('robust performance by DK: %f \t \n', RPdk)
%% Specify IQC multipliers for LTI norm bounded uncertainty

D1 = ss(eye(1));
D2 = ss(tf(0.01,[1 0.01]));
D3 = ss(tf(0.0464,[1 0.0464]));
D4 = ss(tf(0.2152,[1 0.2152]));
D5 = ss(tf(1,[1 1]));

Psis{1} = blkdiag(D1,D1,D1,D1);

Psis{2} = blkdiag(D1,D2,D1,D2);

Psis{3} = blkdiag(D2,D1,D2,D1);

Psis{4} = blkdiag(D1,D3,D1,D3);

Psis{5} = blkdiag(D3,D1,D3,D1);

% Psis{6} = blkdiag(D1,D4,D1,D4);
% 
% Psis{7} = blkdiag(D4,D1,D4,D1);
% 
% Psis{8} = blkdiag(D1,D5,D1,D5);
% 
% Psis{9} = blkdiag(D5,D1,D5,D1);

%% Initialization

szDelta = [2 2];
szK = [2 2];

szG = size(Gic);
if numel(szG) == 2
    szG = [szG 1];
end
nmod = szG(3); % # of models in the G array.

szU = sum(szDelta,1);
nw = szU(1);   % # of outputs of Delta (uncertaint inputs to G)
nv = szU(2);   % # of inputs to Delta

nu = szK(1);    % # of outputs of K (control inputs)
ny = szK(2);    % # of inputs of K (measurements)

nd = szG(2) - nw - nu ;   % # of disturbances
ne = szG(1) - nv - ny ;   % # of errors

% n IQCs
niqcs = numel(Psis);

% lambda for n IQCs
lam = zeros(niqcs,1);
lam(1) = 1;

M = blkdiag(eye(nv),-eye(nw));

% stopping tolerence
maxiter = 25;

gamold = inf;
gamtol = 0.05;

%% Iterations

tic;
for ii = 1:maxiter
%     ii = 2;
    fprintf('\n Iteration: %i \n', ii);
    
    % create a single IQC from n IQCs
    if ii ~= 1
        Pi = ss(zeros(nv + nw,nv + nw));
        Scl = blkdiag(1/gamold*eye(nv),eye(nw));
        for jj = 1:niqcs
            Pi = Pi + lam(jj)*Scl*Psis{jj}'*M*Psis{jj}*Scl;
        end
    end
    % facotorization of the single IQC
    if ii == 1
        Psi = Psis{1};
    else
        [Psi,M] = jspec(Pi); 
        invScl = blkdiag(gamold*eye(nv),eye(nw));
        Psi = invScl*Psi;
    end

    % create Psi_plus by inverting w/w_tilde channels
    Apsi = Psi.a;
    Bpsi = Psi.b;
    Cpsi_vtil = Psi.c(1:nv,:);
    Cpsi_wtil = Psi.c(nv+1:nv+nw,:);
    Dpsi_vtil = Psi.d(1:nv,:);
    Dpsi_wtilv = Psi.d(nv+1:nv+nw,1:nv);
    Dpsi_wtilw = Psi.d(nv+1:nv+nw,nv+1:nv+nw);    

    nxpsi = size(Apsi,1);
    
    Psi_mtx = [Apsi, Bpsi; Cpsi_vtil, Dpsi_vtil; zeros(nw,nxpsi+nv), eye(nw)];

    T = [eye(nxpsi), zeros(nxpsi,nw+nv); zeros(nv,nxpsi+nw), eye(nv); 
            -Dpsi_wtilw\Cpsi_wtil, inv(Dpsi_wtilw), -Dpsi_wtilw\Dpsi_wtilv];

    Psi_plusmtx = Psi_mtx*T;
    
    Psi_plusa = Psi_plusmtx(1:nxpsi,1:nxpsi);
    Psi_plusb = Psi_plusmtx(1:nxpsi,nxpsi+1:nxpsi+nw+nv);
    Psi_plusc = Psi_plusmtx(nxpsi+1:nxpsi+nv+nw,1:nxpsi);
    Psi_plusd = Psi_plusmtx(nxpsi+1:nxpsi+nv+nw,nxpsi+1:nxpsi+nw+nv);
    
    Psi_plus = ss(Psi_plusa,Psi_plusb,Psi_plusc,Psi_plusd);
    
    
    % create scaled system G_scl by lft of G and Psi_plus 
    
    G_scl = lft(Psi_plus,Gic,nv,nw);

    % nominal LPV synthesis step 
    % this "backs off" the optimal gamma by BackOffFactor = 1.1000
    opt = lpvsynOptions;
    opt.BackOffFactor = 1.2;
    [Krob,GAMrob,INFOrob] = lpvsyneng(G_scl,ny,nu);
   
    fprintf('Gamma controller step: %f \t \n', GAMrob)
    
    % robust performance analysis step
    %get analysis interconnection (replace G by Pic for demo)
    Nic = lft(Gic,Krob);
    
    % options for bisection algorithm
    gamlb = 0;  %lower bound
    
    if ii==1    %upper bound   
        gamub = GAMrob; 
    else
%         gamub = gamold; 
        gamub = GAMrob;
    end

    tol = 0.05;
    
    while gamub-gamlb>tol
        gamtry = (gamlb+gamub)/2;
        
        %run analysis
        [t,info] = lpvrbperf(Nic,szDelta,Psis,gamtry);
        %new bounds for bisection
        if t <= 0
            gamub = gamtry;
        else
            gamlb = gamtry;
        end
    end
    
    gamopt = gamub;
    % run analysis using gamopt
    [t,info] = lpvrbperf(Nic,szDelta,Psis,gamopt);

    fprintf('Gamma analysis step: %f \t \n', gamopt)
    %% save results for next iteration
    lam = info.lam;   
    gamvec(ii) = gamopt;
    
    if gamold-gamopt < gamtol
        break;
    end
    gamold = gamopt;
    
end
toc;








