function [gamma,info] = lpvnorm(G)
% function [gamma,info] = lpvnorm(G)
%
% This function computes a upper bound of the induce L2 norm of G which is an 
% LPV system.  The code makes the following
% assumptions to study the use of parameter-varying IQCs:
%   1) G depends on a single parameter p
%   2) There are no rate bounds on the parameter p
%
% Inputs
%  G: An ne-by-nd-by-nmod array of state-space models that 
%      describe the dynamics of the LPV system G on a parameter grid of 
%      nmod points.
%
% Outputs
%  gamma: Bound on the gain of G
%  info:  Structure with information regarding solution of dissipation
%         inequality. Structure includes the fields:
%    -P:  Symmetric matrix used to construct the storage function.
%


%% Dimensions
szG = size(G);
if numel(szG) == 2
    szG = [szG 1];
end
nmod = szG(3); % # of models in the G array.


nd = szG(2);   % # of disturbances
ne = szG(1);   % # of errors

[A,B,C,D] = ssdata(G);
nx = size(A,1);

setlmis([])


%%
% Initialize LMI and define variables
gsqinv = lmivar(1,[1 1]);
P = lmivar(1,[nx 1]);
gsqI = blkdiag(zeros(nx),eye(nd)); 
%%
% Specify LMIs

% LMI for Dissipation Inequality
cnt = 1;
for k = 1:nmod
    Ak = A(:,:,k);
    Bk = B(:,:,k);
    Ck = C(:,:,k);
    Dk = D(:,:,k);
    
    % Terms involving plant and error signals
    lmiterm([cnt 1 1 P],[eye(nx); zeros(nd,nx)],[Ak Bk],'s');
    lmiterm([-cnt 1 1 0],gsqI);
    lmiterm([cnt 1 1 gsqinv],1,[Ck Dk]'*[Ck Dk]);
    
    cnt = cnt +1;
end

% LMI to enforce P>=0
    lmiterm([-cnt 1 1 P],1,1);
    cnt  = cnt+1;
    
    lmiterm([-cnt 1 1 gsqinv],1,1);
    cnt  = cnt+1;    

%% Solve SDP: min gamsq subject to LMI constraints
lmisys = getlmis;

% minimization specification
ndec = decnbr(lmisys);
c=zeros(ndec,1);
c(1)=-1;

% solution
opt = [0 0 0 0 1];
[copt,xopt]=mincx(lmisys,c,opt);
gamma = sqrt(-1/copt);  

%% Store solution
info.P = [];
if ~isempty(xopt)
    info.P = dec2mat(lmisys,xopt,P);
end
